@@ -1,5 +1,5 @@
 # engine/url.py
-# Copyright (C) 2005-2018 the SQLAlchemy authors and contributors
+# Copyright (C) 2005-2019 the SQLAlchemy authors and contributors
 # <see AUTHORS file>
 #
 # This module is part of SQLAlchemy and is released under
@@ -15,9 +15,12 @@ be used directly and is also accepted directly by ``create_engine()``.
 """
 
 import re
-from .. import exc, util
-from . import Dialect
-from ..dialects import registry, plugins
+
+from .interfaces import Dialect
+from .. import exc
+from .. import util
+from ..dialects import plugins
+from ..dialects import registry
 
 
 class URL(object):
@@ -50,11 +53,19 @@ class URL(object):
 
     """
 
-    def __init__(self, drivername, username=None, password=None,
-                 host=None, port=None, database=None, query=None):
+    def __init__(
+        self,
+        drivername,
+        username=None,
+        password=None,
+        host=None,
+        port=None,
+        database=None,
+        query=None,
+    ):
         self.drivername = drivername
         self.username = username
-        self.password = password
+        self.password_original = password
         self.host = host
         if port is not None:
             self.port = int(port)
@@ -68,22 +79,27 @@ class URL(object):
         if self.username is not None:
             s += _rfc_1738_quote(self.username)
             if self.password is not None:
-                s += ':' + ('***' if hide_password
-                            else _rfc_1738_quote(self.password))
+                s += ":" + (
+                    "***" if hide_password else _rfc_1738_quote(self.password)
+                )
             s += "@"
         if self.host is not None:
-            if ':' in self.host:
+            if ":" in self.host:
                 s += "[%s]" % self.host
             else:
                 s += self.host
         if self.port is not None:
-            s += ':' + str(self.port)
+            s += ":" + str(self.port)
         if self.database is not None:
-            s += '/' + self.database
+            s += "/" + self.database
         if self.query:
             keys = list(self.query)
             keys.sort()
-            s += '?' + "&".join("%s=%s" % (k, self.query[k]) for k in keys)
+            s += "?" + "&".join(
+                "%s=%s" % (k, element)
+                for k in keys
+                for element in util.to_list(self.query[k])
+            )
         return s
 
     def __str__(self):
@@ -96,29 +112,46 @@ class URL(object):
         return hash(str(self))
 
     def __eq__(self, other):
-        return \
-            isinstance(other, URL) and \
-            self.drivername == other.drivername and \
-            self.username == other.username and \
-            self.password == other.password and \
-            self.host == other.host and \
-            self.database == other.database and \
-            self.query == other.query
+        return (
+            isinstance(other, URL)
+            and self.drivername == other.drivername
+            and self.username == other.username
+            and self.password == other.password
+            and self.host == other.host
+            and self.database == other.database
+            and self.query == other.query
+            and self.port == other.port
+        )
+
+    def __ne__(self, other):
+        return not self == other
+
+    @property
+    def password(self):
+        if self.password_original is None:
+            return None
+        else:
+            return util.text_type(self.password_original)
+
+    @password.setter
+    def password(self, password):
+        self.password_original = password
 
     def get_backend_name(self):
-        if '+' not in self.drivername:
+        if "+" not in self.drivername:
             return self.drivername
         else:
-            return self.drivername.split('+')[0]
+            return self.drivername.split("+")[0]
 
     def get_driver_name(self):
-        if '+' not in self.drivername:
+        if "+" not in self.drivername:
             return self.get_dialect().driver
         else:
-            return self.drivername.split('+')[1]
+            return self.drivername.split("+")[1]
 
     def _instantiate_plugins(self, kwargs):
-        plugin_names = util.to_list(self.query.get('plugin', ()))
+        plugin_names = util.to_list(self.query.get("plugin", ()))
+        plugin_names += kwargs.get("plugins", [])
 
         return [
             plugins.load(plugin_name)(self, kwargs)
@@ -132,17 +165,19 @@ class URL(object):
         returned class implements the get_dialect_cls() method.
 
         """
-        if '+' not in self.drivername:
+        if "+" not in self.drivername:
             name = self.drivername
         else:
-            name = self.drivername.replace('+', '.')
+            name = self.drivername.replace("+", ".")
         cls = registry.load(name)
         # check for legacy dialects that
         # would return a module with 'dialect' as the
         # actual class
-        if hasattr(cls, 'dialect') and \
-                isinstance(cls.dialect, type) and \
-                issubclass(cls.dialect, Dialect):
+        if (
+            hasattr(cls, "dialect")
+            and isinstance(cls.dialect, type)
+            and issubclass(cls.dialect, Dialect)
+        ):
             return cls.dialect
         else:
             return cls
@@ -170,7 +205,7 @@ class URL(object):
         """
 
         translated = {}
-        attribute_names = ['host', 'database', 'username', 'password', 'port']
+        attribute_names = ["host", "database", "username", "password", "port"]
         for sname in attribute_names:
             if names:
                 name = names.pop(0)
@@ -197,7 +232,8 @@ def make_url(name_or_url):
 
 
 def _parse_rfc1738_args(name):
-    pattern = re.compile(r'''
+    pattern = re.compile(
+        r"""
             (?P<name>[\w\+]+)://
             (?:
                 (?P<username>[^:/]*)
@@ -211,40 +247,53 @@ def _parse_rfc1738_args(name):
                 (?::(?P<port>[^/]*))?
             )?
             (?:/(?P<database>.*))?
-            ''', re.X)
+            """,
+        re.X,
+    )
 
     m = pattern.match(name)
     if m is not None:
         components = m.groupdict()
-        if components['database'] is not None:
-            tokens = components['database'].split('?', 2)
-            components['database'] = tokens[0]
-            query = (
-                len(tokens) > 1 and dict(util.parse_qsl(tokens[1]))) or None
-            if util.py2k and query is not None:
-                query = dict((k.encode('ascii'), query[k]) for k in query)
+        if components["database"] is not None:
+            tokens = components["database"].split("?", 2)
+            components["database"] = tokens[0]
+
+            if len(tokens) > 1:
+                query = {}
+
+                for key, value in util.parse_qsl(tokens[1]):
+                    if util.py2k:
+                        key = key.encode("ascii")
+                    if key in query:
+                        query[key] = util.to_list(query[key])
+                        query[key].append(value)
+                    else:
+                        query[key] = value
+            else:
+                query = None
         else:
             query = None
-        components['query'] = query
+        components["query"] = query
 
-        if components['username'] is not None:
-            components['username'] = _rfc_1738_unquote(components['username'])
+        if components["username"] is not None:
+            components["username"] = _rfc_1738_unquote(components["username"])
 
-        if components['password'] is not None:
-            components['password'] = _rfc_1738_unquote(components['password'])
+        if components["password"] is not None:
+            components["password"] = _rfc_1738_unquote(components["password"])
 
-        ipv4host = components.pop('ipv4host')
-        ipv6host = components.pop('ipv6host')
-        components['host'] = ipv4host or ipv6host
-        name = components.pop('name')
+        ipv4host = components.pop("ipv4host")
+        ipv6host = components.pop("ipv6host")
+        components["host"] = ipv4host or ipv6host
+        name = components.pop("name")
         return URL(name, **components)
     else:
         raise exc.ArgumentError(
-            "Could not parse rfc1738 URL from string '%s'" % name)
+            "Could not parse rfc1738 URL from string '%s'" % name
+        )
 
 
 def _rfc_1738_quote(text):
-    return re.sub(r'[:@/]', lambda m: "%%%X" % ord(m.group(0)), text)
+    return re.sub(r"[:@/]", lambda m: "%%%X" % ord(m.group(0)), text)
 
 
 def _rfc_1738_unquote(text):
@@ -252,7 +301,7 @@ def _rfc_1738_unquote(text):
 
 
 def _parse_keyvalue_args(name):
-    m = re.match(r'(\w+)://(.*)', name)
+    m = re.match(r"(\w+)://(.*)", name)
     if m is not None:
         (name, args) = m.group(1, 2)
         opts = dict(util.parse_qsl(args))
